import com.vividsolutions.jts.io.*;
import com.vividsolutions.jts.geom.*;
import com.vividsolutions.jts.noding.*;
import com.vividsolutions.jts.noding.snapround.*;
import java.util.*; // for Collection, I guess
import java.io.*; 

public class ScaledNoderTest { 

   public static void main(String[] args)
         throws ParseException, java.io.IOException
   { 
      ScaledNoderTest snt = new ScaledNoderTest();
      //snt.test1(args);
      snt.test2(args);
   }

   public void test1(String[] args) throws ParseException {

      String wkti = new String( "LINESTRING(0 0, 10 0,5 5, 5 -5)" );

      ArrayList<SegmentString> nodable = new ArrayList<SegmentString>();
      getSegmentStrings(wkti, nodable);

      PrecisionModel pm = new PrecisionModel(1.0); // fixed as well
      MCIndexSnapRounder inoder = new MCIndexSnapRounder(pm);
      ScaledNoder noder = new ScaledNoder(inoder, 1e5); 

      noder.computeNodes(nodable);
      Collection noded = noder.getNodedSubstrings();

      String wktobt = toWKT(noded);

      String wktexp = "MULTILINESTRING ((0 0, 5 0), (5 0, 10 0), (10 0, 5 5), (5 5, 5 0), (5 0, 5 -5))";

      if ( wktobt.equals( wktexp )  ) {
         System.out.println("PASS: ScaledNoder with scale 1e5 and simple noding");
      } else {
         System.out.println("FAIL: ScaledNoder with scale 1e5 and simple noding");
         System.out.println(" Expected: [" + wktexp + "]");
         System.out.println(" Obtained: [" + wktobt + "]");
      }
    }

   private static String readFileAsString(String filePath)
      throws java.io.IOException
   {
    byte[] buffer = new byte[(int) new File(filePath).length()];
    BufferedInputStream f = null;
    try {
        f = new BufferedInputStream(new FileInputStream(filePath));
        f.read(buffer);
    } finally {
        if (f != null) try { f.close(); } catch (IOException ignored) { }
    }
    return new String(buffer);
   }

   private void printSegmentStrings(Collection col)
   {
      //System.out.println(toWKT(col));
      System.out.println(toHEXWKB(col));
   }

   public void test2(String[] args) throws ParseException, java.io.IOException {

      //String wkti = readFileAsString("/tmp/diffnoding.wkt");
      //String wkti = readFileAsString("/tmp/bufrob4.wkt");
      //String wkti = readFileAsString("/tmp/bufrob3_5.wkt");
      String wkti = readFileAsString("/tmp/bufrob3_5_5.wkt");

      ArrayList<SegmentString> nodable = new ArrayList<SegmentString>();
      getSegmentStrings(wkti, nodable);

      PrecisionModel pm = new PrecisionModel(1.0); // fixed as well
      MCIndexSnapRounder inoder = new MCIndexSnapRounder(pm);
      ScaledNoder noder = new ScaledNoder(inoder, 1e5); 

      noder.computeNodes(nodable);
      Collection noded = noder.getNodedSubstrings();

      //int expedges = 12420;
      //int expedges = 1058; // bufrob2
      //int expedges = 333; // bufrob3
      int expedges = 147; // bufrob3_5
      //int expedges = 9; // bufrob4
      if ( noded.size() == expedges ) {
         System.out.println("PASS: ScaledNoder with scale 1e5 and hard noding");
      } else {
         System.out.println("FAIL: ScaledNoder with scale 1e5 and hard noding");
         System.out.println(" Expected: [" + expedges + "]");
         System.out.println(" Obtained: [" + noded.size() + "]");
      }

      printSegmentStrings(noded);
/*
      String wktobt = toWKT(noded);

      String wktexp = "MULTILINESTRING ((0 0, 5 0), (5 0, 10 0), (10 0, 5 5), (5 5, 5 0), (5 0, 5 -5))";

      if ( wktobt.equals( wktexp )  ) {
         System.out.println("PASS: ScaledNoder with scale 1e5 and hard noding");
      } else {
         System.out.println("FAIL: ScaledNoder with scale 1e5 and hard noding");
         System.out.println(" Expected: [" + wktexp + "]");
         System.out.println(" Obtained: [" + wktobt + "]");
      }
*/
    }

   private GeometryFactory gf_ = new GeometryFactory();

   private String toWKT(Collection noded)
   {
      Geometry g = getGeometry(noded);
      WKTWriter w = new WKTWriter();
      return w.write(g);
   }

   private String toHEXWKB(Collection noded)
   {
      Geometry g = getGeometry(noded);
      WKBWriter w = new WKBWriter();
      byte wkb[] = w.write(g);
      String wkbHex = WKBWriter.toHex(wkb);
      return wkbHex;
   }

   private Geometry getGeometry(Collection nodable)
   {
      CoordinateSequenceFactory csf_ = gf_.getCoordinateSequenceFactory();
      ArrayList<LineString> lines = new ArrayList<LineString>();
      for (Iterator i = nodable.iterator(); i.hasNext(); ) {
         SegmentString ss = (SegmentString)i.next();
         Coordinate coords[] = ss.getCoordinates();
         CoordinateSequence cs = csf_.create(coords);
         LineString ls = gf_.createLineString(cs);
         lines.add(ls);
      }
      return gf_.createMultiLineString(gf_.toLineStringArray(lines));
   }

    private void getSegmentStrings(Geometry g, Collection<SegmentString> nodable)
    {
        Coordinate s[] = g.getCoordinates();
        nodable.add( new NodedSegmentString(s, 0) );
    }

   private Geometry readGeometry(String wkt) throws ParseException
   {
      if ( wkt.charAt(0) == '0' || wkt.charAt(0) == '1' ) {
         WKBReader r = new WKBReader();
         byte wkb[] = WKBReader.hexToBytes(wkt);
         return r.read(wkb);
      } else {
         WKTReader r = new WKTReader();
         return r.read(wkt);
      }
   }

    private void getSegmentStrings(String wkt, Collection<SegmentString> nodable)
      throws ParseException
    {
      Geometry g  = readGeometry(wkt);
      getSegmentStrings(g, nodable);
    }

}
